from classo.path_algorithm import solve_path
import numpy as np
import numpy.linalg as LA
    
'''    
Problem    :   min ||Ab - y||^2/sigma + 0.5 sigma + lambda ||b||1 with C.b= 0 and sigma > 0

Dimensions :   A : m*d  ;  y : m  ;  b : d   ; C : k*d

The first function compute a solution of a Lasso problem for a given lambda. The parameters are lam (lambda/lambdamax, \in [0,1]) and pb, which has to be a 'problem_LS type', which is defined bellow in order to contain all the important parameters of the problem. One can initialise it this way : pb = class_of_problem.problem(data=(A,C,y),type_of_algo). We solve the problem without normalizing anything. 
'''    



def algo_Concomitant(pb,lam, compute=True):
    
    pb_type = pb.type                          # ODE, 2prox
    (m,d,k),(A,C,y)  = pb.dim,pb.matrix
    lamb = lam * pb.lambdamax
    sigmax= LA.norm(y) * np.sqrt(2)

    #ODE
    # here we compute the path algo until our lambda, and just take the last beta
    if(pb_type == 'ODE'):       
                         
        (beta1,beta2), (s1,s2), (r1,r2) = solve_path((A,C,y),lam,concomitant = 'fix_lam')
        dr,ds = r1-r2,s1-s2
        teta  = root_2(LA.norm(dr)**2-ds**2 ,np.vdot(dr,r2)-s2*ds,LA.norm(r2)**2-s2**2)
        s     = s1    *teta + s2    *(1-teta)
        beta  = beta1 *teta + beta2 *(1-teta)
        return(beta,s*sigmax)
    

    if(compute): pb.compute_param()
    
    Anorm         = LA.norm(A,'fro')
    tol,regpath   = pb.tol * LA.norm(y)/Anorm,pb.regpath   # tolerance rescaled
    Proj          = proj_c(C,d)   # Proj = I - C^t . (C . C^t )^-1 . C
    AtA, Aty      = pb.AtA, pb.Aty     # Save some matrix products already computed in problem.compute_param() 
    gamma         = pb.gam / (pb.AtAnorm*lam)   # Normalize gamma 
    w,zerod       = lamb *gamma*pb.weights, np.zeros(d) # two vectors usefull to compute the prox of f(b)= sum(wi |bi|)
    mu, c, root   = pb.mu, pb.c, 0.
    Q1,Q2         = QQ(c,A) 
    QA,qy         = Q1.dot(A), Q1.dot(y)
    proj_sigm     = lambda x : max(x,0)
    xs,nu,o,xbar,x= pb.init

 
    #2prox
    if (pb_type == '2prox' ):
        for i in range(pb.N):
            nv_b, nv_s = x + Q1.dot(o) - QA.dot(x) - Q2.dot(x-xbar), (xs+nu)/2
            if (i>0 and LA.norm(b-nv_b)+LA.norm(s-nv_s)/Anorm <2*tol): 
                if (regpath):return(b,(xs,nu,o,xbar,x),s) 
                else :       return(b,s)
                        
            s,b = nv_s, nv_b
            Ab = A.dot(b)
            p1,p2,root = prox_phi_1(xs,2*Ab-o-y,gamma/c,root)
            sup = [ proj_sigm(nu)-s , p1-s , p2 + y - Ab , prox(2*b-xbar,w,zerod)-b, Proj.dot(2*b-x)-b]
            xs,nu,o,xbar,x = xs+mu*sup[0] ,  nu+mu*sup[1] ,  o+mu*sup[2] ,  xbar+mu*sup[3] ,  x+mu*sup[4]
            if (LA.norm(b)+LA.norm(s)>1e6): 
                print('DIVERGENCE')
                return(b,s)
        print('NO CONVERGENCE')
        return(b,'slow',s)

    
    print('none of the cases ! ')
    

    
    
    
    
    
    
    
    
'''
This function compute the the solution for a given path of lam : by calling the function 'algo' for each lambda with warm start, or wuth the method ODE, by computing the whole path thanks to the ODE that rules Beta and the subgradient s, and then to evaluate it in the given finite path.  
'''
    
def pathalgo_Concomitant(pb,path):
    n = pb.dim[0]
    BETA,SIGMA,tol = [],[],pb.tol

    
    
    if(pb.type=='ODE'):
        y=pb.matrix[2]
        sigmax= LA.norm(y) * np.sqrt(2)
        X,LAM,R = solve_path(pb.matrix,path[-1],concomitant = 'path')
        beta2,l2,r2,j = X[0],path[0]+0.1,-y/LA.norm(y),0
        for lam in path: 
            while (LA.norm(r2)<l2/lam): beta1,l1,r1,beta2,l2,r2,j = beta2,l2,r2,X[j],LAM[j],R[j],j+1
            s1,s2 = l1/lam,l2/lam
            dr,ds = r1-r2,s1-s2
            teta  = root_2(LA.norm(dr)**2-ds**2 ,np.vdot(dr,r2)-s2*ds,LA.norm(r2)**2-s2**2)
            SIGMA.append((s1*teta + s2*(1-teta))*sigmax)
            BETA.append(beta1 *teta + beta2 *(1-teta))
            
        return(BETA,SIGMA)

    
    save_init = pb.init   
    pb.regpath = True
    pb.compute_param()
    for lam in path:
        X = algo_Concomitant(pb,lam,compute=False)     
        BETA.append(X[0])
        SIGMA.append(X[-1])
        pb.init = X[1]    
        if(sum([ (abs(X[0][i])>1e-2) for i in range(len(X[0])) ])>=n):
                pb.init = save_init
                BETA = BETA + [BETA[-1]]*(len(path)-len(BETA))
                pb.regpath = False
                print('stop the path because number of active param reach n')
                return(BETA,SIGMA)
            
    pb.init = save_init
    pb.regpath = False
    return(BETA,SIGMA)








'''
Class of problem : we define a type, which will contain as keys, all the parameters we need for a given problem.
'''


class problem_Concomitant :
    
    def __init__(self,data,algo):
        self.N = 500000
        
        if(len(data)==3):
            (A,C,y) = data
            self.dim = (A.shape[0],A.shape[1],C.shape[0])
            e = self.dim[0]
            A,y = A/np.sqrt(2*e),y/np.sqrt(2*e)
            self.matrix = (A,C,y)
        
        elif(len(data)==5):
            (A,C,sol), self.dim = generate_random(data), data[:3]
            self.sol,y = sol, A.dot(sol)+np.random.randn(data[0])*data[-1]
            self.matrix = (A,C,y)

        
        (m,d,k) = self.dim
        self.weights = np.ones(d)
        
        self.tol = 1e-6
          
        self.regpath = False
        self.name = algo + ' Concomitant'
        self.type = algo          # type of algorithm used
        
        self.proj_sigm = lambda x : max(x,0)
        self.mu = 1.95

        self.gam        = np.sqrt(self.dim[1])
        self.Aty        = (A.T).dot(y)
        self.lambdamax  = np.sqrt(2)*LA.norm(self.Aty,np.infty)/LA.norm(y)
        sigmax          = np.sqrt(2)*LA.norm(y)
        self.init       = 1.,1.,np.zeros(m), np.zeros(d), np.zeros(d)    
            
            
            
    def compute_param(self):
        (A,C,y) = self.matrix
        m,d,k = self.dim
        self.c = (d/LA.norm(A,2))**2  # parameter for Concomitant problem : the matrix is scaled as c*A^2 
        
        self.AtA        =(A.T).dot(A)
        self.Cnorm      = LA.norm(C,2)**2
        self.AtAnorm    = LA.norm(self.AtA,2)
        


        



'''
Functions used in the algorithms, modules needed : 
import numpy as np
import numpy.linalg as LA
from .../class_of_problem import problem
'''


# compute the prox of the function : f(b)= sum (wi * |bi| )
def prox(b,w,zeros): return(np.minimum(b+w,zeros)+np.maximum(b-w,zeros)) 

# Compute I - C^t (C.C^t)^-1 . C : the projection on Ker(C)
def proj_c(M,d):
    if (LA.matrix_rank(M)==0):  return(np.eye(d))
    return(np.eye(d)-LA.multi_dot([M.T,np.linalg.inv(M.dot(M.T) ),M]) )


# Compute the real positive root of a polynomial of degree 3 in the form : X^3 + a*X - b with Newton method and a warm start (for Comcomitant problem)
def calc_Newton(a,b,root):
    er = -b
    while (abs(er)>1e-6): 
        root= root-er/(3*root**2+a)
        er = root**3 + a*root-b
    return(root)


def QQ(coef,A): return(coef*(A.T).dot(LA.inv(2*np.eye(A.shape[0])+coef*A.dot(A.T))),LA.inv(2*np.eye(A.shape[1])+coef*(A.T).dot(A)))    


# Return the cost function of some Beta for a given Lasso problem : L_LS = ||y-Ab||2^2  + lambda* ||b||1
def L_LS(pb,lam,sol): return(LA.norm( pb.matrix[0].dot(sol) - pb.matrix[2] )**2 + lam*pb.lambdamax * LA.norm(sol,1))


# Return the prox operator for the function phi = ||(y-Ab)||2^2 /sigma + 1/2 * sigma with the warm start on the computation of the root    
def prox_phi_1(sig,u,gamma,warm_start):
    l2u = LA.norm(u)
    Po0 = 4*gamma*sig + l2u**2
    if (Po0 < 2*gamma**2): return(0.,0.,0.)
    else : 
        root=calc_Newton((4*sig/gamma+6.),8*l2u/gamma,warm_start)
        return(sig+0.5*gamma*(0.5*root**2-1.),u * (1-gamma*root/l2u),root)


    
# Return the positive root in [0,1] of the polynomial aX^2 + 2bX + c if it exists
def root_2(a,b,c):
    root=-(np.sqrt(b**2-a*c)+b)/a
    if (root< 0 or root > 1) : return(0)
    return(root)








def generate_random(dim):                # Function to generate random A, y, and C with given dimensions 
        (m,d,k,d_nonzero,sigma) = dim
        A, sol, sol_reduc= np.random.randn(m, d),np.zeros(d), np.random.rand(d_nonzero)
        if (k==0):
            C , list_i = np.zeros((1,d)), np.random.randint(d, size=d_nonzero)
            sol[list_i]=sol_reduc
        else:
            rank1,rank2 = 0,0
            while (rank1 !=k) :        
                C = np.random.randint(low=-1,high=2, size=(k, d))
                rank1 = LA.matrix_rank(C)
            while (rank2!=k):
                list_i = np.random.randint(d, size=d_nonzero)
                C_reduc = np.array([C.T[i] for i in list_i]).T
                rank2 = LA.matrix_rank(C_reduc)
            proj = proj_c(C_reduc,d_nonzero).dot(sol_reduc)
            sol[list_i]=proj
            return(A,C,sol)



def courbes_gamma(lG,pb,lam):
    nG, C = len(lG), pb.matrix[1]
    T,F,D = np.zeros(nG), np.zeros(nG), np.zeros(nG)
    for j in range(nG) :
        pb.gam = lG[j]
        X = algo(pb,lam)
        if (X[1]==False): break
        T[j], F[j], D[j] = X[2], L_LS(pb,lam,X[0]), LA.norm(C.dot(X[0]))
    plt.figure(figsize=(14,7), dpi=80),plt.plot(lG,T, label = pb.name)
    plt.xlabel("gamma"),plt.ylabel("Running time"), plt.legend(), plt.show()
    
    plt.figure(figsize=(14,7), dpi=80),plt.plot(lG,F, label = pb.name)
    plt.xlabel("gamma"),plt.ylabel("Value of minimizing f"), plt.legend(), plt.show()    
    
    plt.figure(figsize=(14,7), dpi=80),plt.plot(lG,D, label = pb.name)
    plt.xlabel("gamma"),plt.ylabel("Violation of constraints"), plt.legend(), plt.show() 


colo = ['red','orange','g','b','pink','y','c','m','purple','yellowgreen','silver','coral','plum','lime']
colo = colo*100
import numpy as np
import numpy.linalg as LA
import matplotlib.pyplot as plt
import pandas as pd



def random_data(m,d,d_nonzero,k,sigma,zerosum=False):
    A= np.random.randn(m,d)
    while True :        
        C = np.random.randint(low=-1,high=1, size=(k, d))
        if (LA.matrix_rank(C)==k): break
    
    if (zerosum): C,k = np.ones((1,d)),1
        
    while True:
        sol, sol_reduc = np.zeros(d), np.random.rand(d_nonzero)
        list_i = np.random.randint(d, size=d_nonzero)
        C_reduc = np.array([C.T[i] for i in list_i]).T
        if (LA.matrix_rank(C_reduc)<k): continue
        proj = proj_c(C_reduc,d_nonzero).dot(sol_reduc)
        for i in range(len(list_i)):
            sol[list_i[i]]=proj[i]
        break
    y = A.dot(sol)+np.random.randn(m)*sigma
    return((A,C,y),sol)

def name(tol,string): return(string+'  tol={:.0e}'.format(float(tol)))   

def compare(L,show,lam):
    str1 = 'Execution time : \n \n'
    str2 = '\n Iterations : \n \n'
    str3 = '\n L2-difference between : \n \n'
    str4 = '\n For concomitant algorithms : \n \n'
    for i in range(len(L)) :
        x,niter,dt = L[i][0][:3]
        name = L[i][1]
        if (show=='sep' or show==True):
            plt.title('lam = '+str(lam))
            plt.bar(range(len(x)),x,label=name,color=colo[i])
            plt.legend()
            if (show=='sep'): plt.show()
        if (niter != 0):
            str1+= name+' :'+str(round(dt,5)) + '\n'
            str2+= name+' :'+str(niter)+ '\n'
        for j in range (i+1,len(L)): str3+= name + ' and '+L[j][1] +'             : ' + str(round(LA.norm((L[j][0][0]-x)),4)) + '\n'
    plt.show()
    printsigma =False
    for i in range(len(L)) :
        if (len(L[i][0])==4): 
            printsigma = True
            str4+=' Sigma '+L[i][1]+'  :'+str(round(L[i][0][3],3))
 
    print(str1)
    print(str2)
    if (len(L)>1): print(str3)
    if (printsigma): print(str4)
    
    
def influence(sol,top):
    l_influence = []
    for j in range(top):
        maxi = 0.
        for i in range(len(sol)): 
            if ( not (i in l_influence) and (abs(sol[i])>=maxi)): maxi,argmaxi = abs(sol[i]), i
        l_influence.append(argmaxi)
    return(l_influence)

def plot_influence(labels,l_influence):
    print("influent parameters  : \n \n")
    for beta in l_influence:
        string=''
        for colo in range(1,7):
            string+= str(labels.loc[beta+1,colo])
        print(string + '\n')


def plot_betai(labels,l_index,path,BETA):
    for i in range(len(l_index)) :
        leg = 'index = '+str(l_index[i]+1)
        if not (type(labels)==bool): leg+='  '+ str(labels.loc[l_index[i]+1,6])
        if (i>10): plt.plot(path,[BETA[ilam][l_index[i]] for ilam in range(len(path))],color=colo[l_index[i]])
        else:     plt.plot(path,[BETA[ilam][l_index[i]] for ilam in range(len(path))],label=leg,color=colo[l_index[i]])

def normalize(lb,lna,ly):
    for j in range(len(lb[0])):
        lb[:,j] =  lb[:,j]*ly/lna[j]
    return(lb)     
        
        
        
def old_normalize(lb,lna,ly):
    lbetaprime = []
    for beta in lb:
        lbetaprime.append( np.array([ly*beta[j]/(lna[j]) for j in range(len(beta))]) )
    return(lbetaprime)

def denorm(B,lna,ly): return(np.array([ly*B[j]/(np.sqrt(len(B))*lna[j]) for j in range(len(B))]) ) 
    

def affichage(LISTE_BETA,path,title=' ',labels=False,pix=False):
    naffichage = len(LISTE_BETA[0])//4
    l_index  = influence(LISTE_BETA[-1],naffichage)
    plt.figure(figsize=(10,3), dpi=80)
    if (pix=='path'): plt.plot(path,[0]*len(path),'r+')
    plot_betai(labels,l_index,path,LISTE_BETA)
    plt.title(title)
    plt.legend(loc=4, borderaxespad=0.)
    plt.xlabel("lam")
    plt.ylabel("Betai")
    #plt.savefig('path')
    if (type(pix)==bool and pix==True):
        plt.matshow([[(abs(LISTE_BETA[i][j])>1e-2) for i in range(len(LISTE_BETA))] for j in range(len(LISTE_BETA[0]))])
        plt.show()
    
def aff(M):
    (m,n)= M.shape
    for i in range(m):
        string =  ''
        for j in range(n):
            to_ad=str(round(abs(M[i,j]),3))
            string+=to_ad+' '*(5-len(to_ad))+' |'
        print(string + '//')    
        
        
        
def csv_to_mat(file):
    tab1=pd.read_csv(file)
    return(np.array(tab1)[:,1:])
    

    
    
def rescale(matrix):
    (A,C,y)=matrix
    my = sum(y)/len(y)
    la = [LA.norm(A[:,j]) for j in range(len(A[0]))]
    ly = LA.norm(y-my*np.ones(len(y)))
    An = np.array([[A[i,j]/(la[j])  for j in range(len(A[0]))] for i in range(len(A))])
    yn = np.array([ (y[j]-my)/ly for j in range(len(y)) ])
    Cn = np.array([[C[i,j]*ly/la[j] for j in range(len(A[0]))] for i in range(len(C))])
    return((An,Cn,yn),(la,ly,my))


def hub(r,rho) : 
    h=0
    for j in range(len(r)):
        if(abs(r[j])<rho): h+=r[j]**2
        elif(r[j]>0)     : h+= (2*r[j]-rho)*rho
        else             : h+= (-2*r[j]-rho)*rho
    return(h)
def L_LS(A,y,lamb,x): return(LA.norm( A.dot(x) - y )**2 + lamb * LA.norm(x,1))
def L_conc(A,y,lamb,x): return(LA.norm( A.dot(x) - y ) + np.sqrt(2)*lamb * LA.norm(y,1))
def L_H(A,y,lamb,x,rho): return(hub( A.dot(x) - y , rho) + lamb * LA.norm(x,1))


# Compute I - C^t (C.C^t)^-1 . C : the projection on Ker(C)
def proj_c(M,d):
    if (LA.matrix_rank(M)==0):  return(np.eye(d))
    return(np.eye(d)-LA.multi_dot([M.T,np.linalg.inv(M.dot(M.T) ),M]) )

from scipy.special import erfinv
def model_selection(n,p):
    x=0.
    dx = 0.1
    for i in range(10):
        bo = True
        while bo :
            x    += dx
            f     = erfinv(1-x)
            xp    = 8/p * (f**4+f**2)
            bo    = (xp>x)           
        x = x-dx
        dx = dx/10
    return(f*2/np.sqrt(n))



from classo.path_algorithm_huber import solve_huber_path
import cvxpy as cp
import numpy as np
import numpy.linalg as LA
from classo.solve_LS import problem_LS, algo_LS
    
'''    
Problem    :   min h_rho(Ab - y) + lambda ||b||1 with C.b= 0 <=>   min ||Ab - y - r*o||^2 + lambda ||b,o||1 with C.b= 0, o in R^m
                                                                                        r = lambda / 2rho
Dimensions :   A : m*d  ;  y : m  ;  b : d   ; C : k*d

The first function compute a solution of a Lasso problem for a given lambda. The parameters are lam (lambda/lambdamax, \in [0,1]) and pb, which has to be a 'problem_LS type', which is defined bellow in order to contain all the important parameters of the problem. One can initialise it this way : pb = class_of_problem.problem(data=(A,C,y),type_of_algo). We solve the problem without normalizing anything. 
'''    


def algo_Huber(pb,lam, compute=True):
    
    pb_type = pb.type   # ODE, FB, cvx
    
    
    (m,d,k),(A,C,y)  = pb.dim,pb.matrix        
    lamb, rho = lam * pb.lambdamax,  pb.rho
    
    
    #ODE
    # here we compute the path algo until our lambda, and just take the last beta
    
    if(pb_type == 'ODE'):                         
        beta = solve_huber_path((A,C,y), lam,rho)[0]
        return(beta[-1])
    
    tol = pb.tol * LA.norm(y)/LA.norm(A,'fro')  # tolerance rescaled
    regpath = pb.regpath
    
    #cvx 
    # call to the cvx function of minimization
    if (pb_type == 'cvx'):
        x = cp.Variable(d)
        objective, constraints = cp.Minimize(cp.sum(cp.huber(A*x - y, rho))+lamb*cp.norm(x, 1)), [C*x == 0]
        prob = cp.Problem(objective, constraints)
        result = prob.solve(warm_start=regpath,eps_abs= tol)
        if (regpath): return(x.value,True) 
        return(x.value)
     
    if(compute): pb.compute_param()
    tau,r,Proj, AtA, Aty = pb.tauN, lamb/(2*rho),  proj_c(C,d), pb.AtA, pb.Aty  
    gamma                 = pb.gam / (2*(pb.AtAnorm+r**2))   
    t                     = lamb * gamma
    w,tm,zerom,zerod      = t*pb.weights,t*np.ones(m), np.zeros(m), np.zeros(d) 
    o,xbar,x,v = pb.init
    # vectors usefull to compute the prox of f(b)= sum(wi |bi|)
    
    #FORWARD BACKWARD
    if (pb_type=='FB'):
        
        for i in range(pb.N):                        
            grad = (AtA.dot(x)-Aty)                  
            v = v + tau*C.dot(xbar)             
            S = x - 2*gamma*grad - 2*gamma*r*(A.T).dot(o) - (C.T).dot(v) 
            o = prox(o*(1-2*gamma*r**2) + 2*gamma*r*(y-A.dot(x)),tm,zerom)
            p = prox(S,w,zerod)             
            nw_x = Proj.dot(p)                   
            eps = nw_x - x                      
            xbar= p + eps                     

            if (i%10==2 and LA.norm(eps)<tol):      # 0.6
                if (regpath): return(x,(o,xbar,x,v)) 
                else :        return(x) 
            x= nw_x 
            if (LA.norm(x)>1e10): 
                print('DIVERGENCE')
                return(x)
        print('NO CONVERGENCE')
        return(x)
     
        
    # NO PROJ
    
    if (pb_type == 'Noproj'):     # y1 --> S ; p1 --> p . ; p2 --> y2
        for i in range(pb.N):    
            grad = (AtA.dot(x)-Aty)
            
            S1 = x - 2*gamma*grad - 2*gamma*r*(A.T).dot(o) - (C.T).dot(v)
            S2 = o*(1-2*gamma*r**2) + 2*gamma*r*(y-A.dot(x))
            
            p1 = prox(S1,w,zerod)
            p2 = prox(S2,tm,zerom)
            
            v   = v + tau*C.dot(p1)
            v2  = v + tau*C.dot(x)
            
            eps1 = p1 + 2*gamma*(Aty-AtA.dot(p1)-r*A.T.dot(o)) - C.T.dot(v2)-S1
            eps2 = p2 + 2*r*gamma*(y-r*p2-A.dot(x)) -S2
            
            x = x + eps1
            o = o + eps2
            
            if (i>0 and LA.norm(eps1)+LA.norm(eps2)<tol): 
                if (regpath): return(x,(o,xbar,x,v)) 
                else :        return(x) 
            
            if (LA.norm(x)+LA.norm(o)+LA.norm(v)>1e6): 
                print('DIVERGENCE')
                return(x)
        print('NO CONVERGENCE')
        return(x)
    
    
    
    
    # 2 prox :
    
    if (pb_type == '2prox'):
        Ahuber = np.concatenate((A,r*np.eye(len(A))),axis = 1 )
        Chuber = np.concatenate((C,np.zeros((len(C),len(y)))),axis = 1 )
        matrices_huber = (Ahuber,Chuber,y)
        prob = problem_LS(matrices_huber,'2prox')
        prob.regpath = regpath
        if (len(pb.init)==3): prob.init = pb.init
        if not (regpath): return(algo_LS(prob,lamb/prob.lambdamax)[:d])
        x , warm_start = algo_LS(prob,lamb/prob.lambdamax)
        return(x[:d],warm_start)
    
    

    
'''
This function compute the the solution for a given path of lam : by calling the function 'algo' for each lambda with warm start, or wuth the method ODE, by computing the whole path thanks to the ODE that rules Beta and the subgradient s, and then to evaluate it in the given finite path.  
'''
    
def pathalgo_Huber(pb,path):
    n = pb.dim[0]
    BETA,tol = [],pb.tol
    if(pb.type == 'ODE'):
        X,sp_path = solve_huber_path(pb.matrix,path[-1],pb.rho)
        i=0
        for lam in path:
            while (lam<sp_path[i+1]): i+=1
            teta = (sp_path[i]-lam)/(sp_path[i]-sp_path[i+1])
            BETA.append(X[i]*(1-teta)+X[i+1]*teta)
        return(BETA)
    
    save_init = pb.init   
    pb.regpath = True
    pb.compute_param()
    for lam in path:
        X = algo_Huber(pb,lam,compute=False)  
        BETA.append(X[0])
        pb.init = X[1]
        if(sum([ (abs(X[0][i])>1e-2) for i in range(len(X[0])) ])>=n):
                pb.init = save_init
                BETA = BETA + [BETA[-1]]*(len(path)-len(BETA))
                pb.regpath = False
                print('stop the path because number of active param reach n')
                return(BETA)
            
    pb.init = save_init
    pb.regpath = False
    return(BETA)








'''
Class of problem : we define a type, which will contain as keys, all the parameters we need for a given problem.
'''


class problem_Huber :
    
    def __init__(self,data,algo,rho_to_normalise):
        self.N = 500000
        
        if(len(data)==3):self.matrix, self.dim = data, (data[0].shape[0],data[0].shape[1],data[1].shape[0])
        
        elif(len(data)==5):
            (A,C,sol), self.dim = generate_random(data), data[:3]
            self.sol,y = sol, A.dot(sol)+np.random.randn(data[0])*data[-1]
            self.matrix = (A,C,y)
        
        (m,d,k) = self.dim
        rho = rho_to_normalise * LA.norm(self.matrix[2],np.infty)/np.sqrt(m)
        self.weights = np.ones(d)
        self.init = np.zeros(m), np.zeros(d), np.zeros(d), np.zeros(k)
        self.tol = 1e-6
        self.regpath = False
        self.name = algo + ' Huber'
        self.type = algo        
        self.rho = rho
        self.gam = 1.
        self.tau = 0.5         # equation for the convergence of Noproj and LS algorithms : gam + tau < 1
        self.lambdamax = 2*LA.norm(self.matrix[0].T.dot(h_prime(self.matrix[2],rho)),np.infty)
            
        
           
            
            
            
            
    def compute_param(self):
        (A,C,y) = self.matrix
        m,d,k = self.dim
        self.c = (d/LA.norm(A,2))**2  # parameter for Concomitant problem : the matrix is scaled as c*A^2 
        

        self.AtA        =(A.T).dot(A)
        self.Aty        = (A.T).dot(y)
        self.Cnorm      = LA.norm(C,2)**2
        self.tauN       = self.tau/self.Cnorm
        self.AtAnorm    = LA.norm(self.AtA,2)
        



'''
Functions used in the algorithms, modules needed : 
import numpy as np
import numpy.linalg as LA
from .../class_of_problem import problem
'''


# compute the prox of the function : f(b)= sum (wi * |bi| )
def prox(b,w,zeros): return(np.minimum(b+w,zeros)+np.maximum(b-w,zeros)) 

# Compute I - C^t (C.C^t)^-1 . C : the projection on Ker(C)
def proj_c(M,d):
    if (LA.matrix_rank(M)==0):  return(np.eye(d))
    return(np.eye(d)-LA.multi_dot([M.T,np.linalg.inv(M.dot(M.T) ),M]) )

# Compute the derivative of the huber function, particulary useful for the computing of lambdamax 
def h_prime(y,rho):
    m = len(y)
    lrho = rho*np.ones(m)
    return(np.maximum(lrho,-y)+ np.minimum(y-lrho,0))
        






def generate_random(dim):                # Function to generate random A, y, and C with given dimensions 
        (m,d,k,d_nonzero,sigma) = dim
        A, sol, sol_reduc= np.random.randn(m, d),np.zeros(d), np.random.rand(d_nonzero)
        if (k==0):
            C , list_i = np.zeros((1,d)), np.random.randint(d, size=d_nonzero)
            sol[list_i]=sol_reduc
        else:
            rank1,rank2 = 0,0
            while (rank1 !=k) :        
                C = np.random.randint(low=-1,high=2, size=(k, d))
                rank1 = LA.matrix_rank(C)
            while (rank2!=k):
                list_i = np.random.randint(d, size=d_nonzero)
                C_reduc = np.array([C.T[i] for i in list_i]).T
                rank2 = LA.matrix_rank(C_reduc)
            proj = proj_c(C_reduc,d_nonzero).dot(sol_reduc)
            sol[list_i]=proj
            return(A,C,sol)




from classo.path_algorithm import solve_path
import cvxpy as cp
import numpy as np
import numpy.linalg as LA
    
'''    
Problem    :   min ||Ab - y||^2 + lambda ||b||1 with C.b= 0

Dimensions :   A : m*d  ;  y : m  ;  b : d   ; C : k*d

The first function compute a solution of a Lasso problem for a given lambda. The parameters are lam (lambda/lambdamax, \in [0,1]) and pb, which has to be a 'problem_LS type', which is defined bellow in order to contain all the important parameters of the problem. One can initialise it this way : pb = class_of_problem.problem(data=(A,C,y),type_of_algo). We solve the problem without normalizing anything. 
'''    



def algo_LS(pb,lam, compute=True):
    
    pb_type = pb.type   # ODE, cvx, Noproj, FB, 2prox_old, 2prox

    # ODE
    # here we compute the path algo until our lambda, and just take the last beta
    if(pb_type == 'ODE'):
        BETA = solve_path(pb.matrix, lam)[0]
        return(BETA[-1])
    
    (m,d,k),(A,C,y)  = pb.dim,pb.matrix            
    lamb = lam * pb.lambdamax
    tol = pb.tol * LA.norm(y)/LA.norm(A,'fro')  # tolerance rescaled
    regpath = pb.regpath 
    
    
    #cvx
    # call to the cvx function of minimization
    if (pb_type == 'cvx'):
        lamb = lam*2*LA.norm(A.T.dot(y),np.infty)
        x = cp.Variable(d)
        objective, constraints = cp.Minimize(cp.sum_squares(A*x-y)+ lamb*cp.norm(x, 1)), [C*x == 0]
        prob = cp.Problem(objective, constraints)
        result = prob.solve(warm_start=regpath,eps_abs= tol)
        if (regpath): return(x.value,True) 
        return(x.value)
    
    

    if(compute): pb.compute_param()
    Proj,AtA, Aty = proj_c(C,d), pb.AtA, pb.Aty     # Save some matrix products already computed in problem.compute_param()
    gamma, tau    = pb.gam / (2*pb.AtAnorm),    pb.tauN
    w,zerod       = lamb *gamma*pb.weights, np.zeros(d) # two vectors usefull to compute the prox of f(b)= sum(wi |bi|)


    
    # NO PROJ
    
    if (pb_type == 'Noproj'):     # y1 --> S ; p1 --> p . ; p2 --> y2
        (x,v) = pb.init
        for i in range(pb.N):    
            
            S = x - gamma*(AtA.dot(x)-Aty)*2 - (C.T).dot(v)
            p = prox(S,w,zerod)
            
            y2 = v + tau*C.dot(x)
            v  = v + tau*C.dot(p)
            
            eps = p - gamma*(AtA.dot(p)-Aty)*2 - C.T.dot(y2)-S
            x = x + eps
            
            if (i>0 and LA.norm(eps)<tol): 
                if (regpath): return(x,(x,v)) 
                else :        return(x) 
            
            if (LA.norm(x)+LA.norm(p)+LA.norm(v)>1e6): 
                print('DIVERGENCE')
                return(x)
        print('NO CONVERGENCE')
        return(x)
    

    #FORARD BACKWARD
    
    if (pb_type=='FB'):
        xbar,x,v = pb.init
        for i in range(pb.N):                   
            grad = (AtA.dot(x)-Aty)
            v = v + tau*C.dot(xbar)            
            s = x - 2*gamma*grad - (C.T).dot(v) 
            p = prox(s,w,zerod)              
            nw_x = Proj.dot(p)
            
            eps = nw_x - x                       
            xbar= p + eps                        
            
            
            if (i%10==2 and LA.norm(eps)<tol):      # 0.6
                if (regpath): return(x,(xbar,x,v)) 
                else :        return(x) 
            x= nw_x 
            if (LA.norm(x)>1e10): 
                print('DIVERGENCE')
                return(x)
        print('NO CONVERGENCE')
        return(x)
    
    
    
    gamma             = gamma/(2*lam)
    w                 = w /(2*lam)
    mu,ls, c, root    = pb.mu,[], pb.c, 0.
    



    
     # 2 PROX  
    if (pb_type=='2prox'):
        
        Q1,Q2  = QQ(2*gamma/(mu-1),A)
        QA,qy = Q1.dot(A), Q1.dot(y)
        
        qy_mult = qy*(mu-1)
        
        b,xbar,x = pb.init 
        for i in range(pb.N):
            xbar= xbar + mu*(prox(2*b-xbar,w,zerod)-b)
            x   = x    + mu*(Proj.dot(2*b-x)       -b)

            nv_b = (2-mu)*b
            nv_b = nv_b + qy_mult + Q2.dot(x+xbar- 2*nv_b)
            if (i%2==1 and LA.norm(b-nv_b)<tol): 
                if (regpath):return(b,(b,xbar,x)) 
                else :       return(b)  
    
            b = nv_b

        print('NO CONVERGENCE')
        return(b)  
 
    
#     if (pb_type=='2prox_old'):
        
#         Q1,Q2  = QQ(c,A) 
#         QA,qy = Q1.dot(A), Q1.dot(y)
        
#         qo,xbar,x = pb.init 
#         proxLS = 1/(1+2*gamma/c)
        
#         QA_mult = QA*(2*mu*proxLS-mu)
#         qy_mult = qy*(mu-mu*proxLS)
        
#         for i in range(pb.N):
#             nv_b = x-QA.dot(x)+qo - Q2.dot(x-xbar)
#             if (i%2==1 and LA.norm(b-nv_b)<tol): 
#                 if (regpath):return(b,(qo,xbar,x)) 
#                 else :       return(b)          
#             b = nv_b
             
#             qo = qo*(1-mu*proxLS)+QA_mult.dot(b)+qy_mult
            
#             xbar,x = xbar+mu*(prox(2*b-xbar,w,zerod)-b) ,  x+mu*(Proj.dot(2*b-x)-b)
    
#         print('NO CONVERGENCE')
#         return(b)  
    
'''
This function compute the the solution for a given path of lam : by calling the function 'algo' for each lambda with warm start, or wuth the method ODE, by computing the whole path thanks to the ODE that rules Beta and the subgradient s, and then to evaluate it in the given finite path.  
'''
    
def pathalgo_LS(pb,path,return_sp_path=False):
    n = pb.dim[0]
    BETA,tol = [],pb.tol
    if(pb.type == 'ODE'):
        beta,sp_path = solve_path(pb.matrix,path[-1])
        if (return_sp_path): return(beta,sp_path)
        i=0
        for lam in path:
            while (lam<sp_path[i+1]): i+=1
            teta = (sp_path[i]-lam)/(sp_path[i]-sp_path[i+1])
            BETA.append(beta[i]*(1-teta)+beta[i+1]*teta)
        return(BETA)
    
    
    
    save_init = pb.init   
    pb.regpath = True
    pb.compute_param()
    for lam in path:
        X = algo_LS(pb,lam,compute=False)     
        BETA.append(X[0])
        pb.init = X[1]    
        if(sum([ (abs(X[0][i])>1e-2) for i in range(len(X[0])) ])>=n):
                pb.init = save_init
                BETA = BETA + [BETA[-1]]*(len(path)-len(BETA))
                pb.regpath = False
                print('stop the path because number of active param reach n')
                return(BETA)
            
    pb.init = save_init
    pb.regpath = False
    return(BETA)








'''
Class of problem : we define a type, which will contain as keys, all the parameters we need for a given problem.
'''


class problem_LS :
    
    def __init__(self,data,algo):
        self.N = 500000
        
        if(len(data)==3):self.matrix, self.dim = data, (data[0].shape[0],data[0].shape[1],data[1].shape[0])
        
        elif(len(data)==5):
            (A,C,sol), self.dim = generate_random(data), data[:3]
            self.sol,y = sol, A.dot(sol)+np.random.randn(data[0])*data[-1]
            self.matrix = (A,C,y)
        
        (m,d,k) = self.dim
        
        if(algo=='FB') : self.init = np.zeros(d), np.zeros(d), np.zeros(k)
        elif (algo=='Noproj')         : self.init = np.zeros(d), np.zeros(k)
        else                        : self.init = np.zeros(d), np.zeros(d), np.zeros(d)
        self.tol = 1e-6 
         
        self.weights = np.ones(d)  
        self.regpath = False
        self.name = algo + ' LS'
        self.type = algo          # type of algorithm used
        self.mu = 1.95
        self.Aty= (self.matrix[0].T).dot(self.matrix[2])
        self.lambdamax = 2*LA.norm(self.Aty,np.infty)
        self.gam = 1.
        self.tau = 0.5         # equation for the convergence of Noproj and LS algorithms : gam + tau < 1
        if (algo in ['2prox_old','2prox']): self.gam = self.dim[1]

            

    def compute_param(self):
        (A,C,y) = self.matrix
        m,d,k = self.dim
        self.c = (d/LA.norm(A,2))**2  # parameter for Concomitant problem : the matrix is scaled as c*A^2 

        self.AtA        =(A.T).dot(A)
        self.Cnorm      = LA.norm(C,2)**2
        self.tauN       = self.tau/self.Cnorm
        self.AtAnorm    = LA.norm(self.AtA,2)

        
        
        
        



'''
Functions used in the algorithms, modules needed : 
import numpy as np
import numpy.linalg as LA
from .../class_of_problem import problem
'''


# compute the prox of the function : f(b)= sum (wi * |bi| )
def prox(b,w,zeros): return(np.minimum(b+w,zeros)+np.maximum(b-w,zeros)) 

# Compute I - C^t (C.C^t)^-1 . C : the projection on Ker(C)
def proj_c(M,d):
    if (LA.matrix_rank(M)==0):  return(np.eye(d))
    return(np.eye(d)-LA.multi_dot([M.T,np.linalg.inv(M.dot(M.T) ),M]) )



def QQ(coef,A): return(coef*(A.T).dot(LA.inv(2*np.eye(A.shape[0])+coef*A.dot(A.T))),LA.inv(2*np.eye(A.shape[1])+coef*(A.T).dot(A)))    



def generate_random(dim):                # Function to generate random A, y, and C with given dimensions 
        (m,d,k,d_nonzero,sigma) = dim
        A, sol, sol_reduc= np.random.randn(m, d),np.zeros(d), np.random.rand(d_nonzero)
        if (k==0):
            C , list_i = np.zeros((1,d)), np.random.randint(d, size=d_nonzero)
            sol[list_i]=sol_reduc
        else:
            rank1,rank2 = 0,0
            while (rank1 !=k) :        
                C = np.random.randint(low=-1,high=2, size=(k, d))
                rank1 = LA.matrix_rank(C)
            while (rank2!=k):
                list_i = np.random.randint(d, size=d_nonzero)
                C_reduc = np.array([C.T[i] for i in list_i]).T
                rank2 = LA.matrix_rank(C_reduc)
            proj = proj_c(C_reduc,d_nonzero).dot(sol_reduc)
            sol[list_i]=proj
            return(A,C,sol)




N=10000
import numpy as np
import numpy.linalg as LA


'''
Main function that solve the Least square problem for all lambda using the ODE, which is describe in the paper : "Algorithms for Fitting the Constrained Lasso"
'''



def solve_huber_path(matrices,lamin,rho,n_active=False,return_lmax=False):
    global number_act,idr,Xt,activity,F,beta,s, lam, M,r
    
    (A,C,y)   = matrices
    m,d,k       = len(A),len(A[0]),len(C)

    lam,LAM,beta,BETA,r,activity,idr,F,number_act = 1.,[1.],np.zeros(d),[np.zeros(d)],-y,[False]*d,[False]*k,[True]*m,0

    for j in range(m):
        if (abs(y[j])>rho): F[j]=False    
      
    s         = 2*A.T.dot(h_prime(y,rho))
    lambdamax = LA.norm(s,np.inf)
    s = s/lambdamax        
    AtA = A[F].T.dot(A[F])
    if (k==0): M = 2*AtA
    else : M  = np.concatenate((np.concatenate((2*AtA,C.T),axis=1),np.concatenate(( C,np.zeros((k, k)) ),axis=1)), axis=0)    
        
    
    for i in range(d):
        if (s[i]==1. or s[i]==-1.): 
            activity[i] = True
            number_act +=1
            if(k>0): 
                to_ad = next_idr1(idr,C[:,activity])
                if(type(to_ad)==int): idr[to_ad] = True
    
    Xt = LA.inv(M[activity+idr,:] [:,activity+idr])    # initialise Xt
    for i in range(N) :
        
        up(lambdamax,lamin,A,rho)
        BETA.append(beta), LAM.append(lam)
        if ((type(n_active)==int and number_act>= n_active) or lam == lamin): 
            if(return_lmax):    return(BETA,LAM,lambdamax)
            else:               return(BETA,LAM)
            
    print('no conv')
    if(return_lmax):    return(BETA,LAM,lambdamax)
    else:               return(BETA,LAM)

#function that search the next lambda where something happen, and update the solution Beta
def up(lambdamax,lamin,A,rho):
    global number_act,idr,Xt,activity,F,beta,s,lam,M,r
    
    d=len(activity)
    L = [lam]*d
    D,E = direction(activity,s,M[:len(activity),:][:,:len(activity)],M[d:,:][:,:d],Xt,idr,number_act)  
    for i in range(d):
        bi,di,e,s0 = beta[i],D[i],E[i],s[i]
        if (activity[i]): 
            if (abs(bi*di)>1e-10 and bi*di<0): 
                L[i]=-bi/(di*lambdamax)
        else :  
            if(abs(e-s0)<1e-10 or abs(s0)>1): continue
            if (e>s0): dl = (1+s0)/ (1+e)
            else: dl = (1-s0)/ (1-e)
            L[i] = dl * lam
        
    dlamb = min(min(L),lam,lam-lamin)
    huber_up,ADl = False, A.dot(D)*lambdamax
    j_switch=None
    for j in range(len(r)):
        #     find if there is a  0< dl < dlamb such that F[j] and |r[j]+ADl[j]*dl|>rho
        # or  find if there is a  0< dl < dlamb such that not F[j] |r[j]+ADl[j]*dl|<rho
        if(abs(ADl[j])<1e-6 or abs(r[j])<1e-6 or abs(abs(r[j])-rho)<1e-6): continue
        c=-r[j]/ADl[j]
        dl=(1-rho/abs(r[j]))*c
        if (dl<0): 
            if(c>0):dl=(1+rho/abs(r[j]))*c
            else:  dl = dlamb      
        if(dl<dlamb): 
            huber_up,j_switch,dlamb=True,j,dl
    
    beta,s,r,lam = beta + lambdamax*D * dlamb,E + lam /(lam-dlamb) * (s-E),r+ADl*dlamb,lam-dlamb
    
    
    if(huber_up):
        F[j_switch] = not F[j_switch]
        M[:d,:][:,:d] = 2*A[F].T.dot(A[F])
        Xt            = LA.inv(M[activity+idr,:] [:,activity+idr])
    else:
    # Update matrix inverse, list of rows in C and activity
        for i in range(d):
            if (L[i]<dlamb+1e-10):
                if (activity[i]):
                    activity[i], number_act = False, number_act - 1
                    if(len(M)>d):
                        to_ad = next_idr2(idr,M[d:,:][:,:d][:,activity])
                        if(type(to_ad)==int): idr[to_ad] = False
                else:
                    #x = M[:,activity+idr][i]
                    #al = M[i,i]-np.vdot(x,Xt.dot(x))
                    #if (abs(al)<1e-10): break
                    activity[i],number_act = True, number_act+1
                    if(len(M)>d): 
                        to_ad = next_idr1(idr,M[d:,:][:,:d][:,activity])
                        if(type(to_ad)==int): idr[to_ad] = True
                Xt  = LA.inv(M[activity+idr,:] [:,activity+idr]) 


# Compute the derivatives of the solution Beta and the derivative of lambda*subgradient thanks to the ODE
def direction(activity,s,Mat,C,Xt,idr,number_act):
    if (len(C)==0):
        D,product =np.zeros(len(activity)), Xt[:,:number_act].dot(s[activity])
        D[activity]= product
        return(D,Mat.dot(D))
    D,Dadj=np.zeros(len(activity)),np.zeros(len(C))
    product = Xt[:,:number_act].dot(s[activity])
    D[activity],Dadj[idr]=product[:number_act],product[number_act:]
    E = (Mat.dot(D)+C.T.dot(Dadj))   #D and D2 in Rd with zeros in inactives and E. D is - derivatives
    return(D,E)      


#Upddate a list of constraints which are independant if we restrict the; to the acrive set (C_A has to have independant rows)
def next_idr1(liste,mat):
    if(sum(liste)==len(mat)): return(False)
    if (sum(liste)==0):
        for i in range(len(mat)):
            for j in range(len (mat[0])):
                if not (mat[i,j]==0): return(i)
        return(False)
    Q = LA.qr(mat[liste,:].T)[0]
    for j in range(len(mat)):
        if (not liste[j]) and (  LA.norm(mat[j]-LA.multi_dot([Q,Q.T,mat[j]]))>1e-10 ): return(j)
    return(False) 
def next_idr2(liste,mat):
    if(sum(liste)==0): return(False)
    R = LA.qr(mat[liste,:].T)[1]
    for i in range(len(R)):
        if(abs(R[i,i])<1e-10):
            j,somme = 0, liste[0]
            while(somme<=i):
                j, somme = j+1, somme + liste[j+1]
            return(j)
    return(False)

#Update the invers of a matrix whom we add a line, which is useul to compute the derivatives
def next_inv(Xt,B,al,ligne):
    n=len(Xt)
    Yt = np.zeros((n+1,n+1))
    alpha = 1/al
    B = np.array([B])
    b1 = Xt[:ligne,:][:,:ligne]+ alpha*B[:,:ligne].T.dot(B[:,:ligne])
    b2 = Xt[ligne:,:][:,:ligne]+ alpha*B[:,ligne:].T.dot(B[:,:ligne])
    b4 = Xt[ligne:,:][:,ligne:]+ alpha*B[:,ligne:].T.dot(B[:,ligne:])
    col1 = np.concatenate((b1,-alpha*B[:,:ligne],b2), axis = 0)
    col2 = np.concatenate((b2.T,-alpha*B[:,ligne:],b4), axis = 0)
    col = np.concatenate((-alpha*B[0,:ligne],[alpha],-alpha*B[0,ligne:]), axis = 0)
    return(np.concatenate((col1,np.array([col]).T,col2), axis = 1))


  

'''
Function that solve the concomitant problem for every lambda thanks to the previous function : we firstly compute all the non concomitant Least square problems, then we use it to find sigma and so the solution, using the equation for sigma: 

sigma = || A*B(lambda*sigma) - y ||_2       (where B(lambda) is found thanks to solve_path)

            teta = (sp_path[i]-lam)/(sp_path[i]-sp_path[i+1])
'''
    

    
    
# Compute the derivative of the huber function, particulary useful for the computing of lambdamax 
def h_prime(y,rho):
    m = len(y)
    lrho = rho*np.ones(m)
    return(np.maximum(lrho,-y)+ np.minimum(y-lrho,0))    
    


from time import time
import numpy as np
import matplotlib.pyplot as plt

from classo.solve_LS import problem_LS, algo_LS, pathalgo_LS
from classo.solve_Huber import problem_Huber, algo_Huber, pathalgo_Huber
from classo.solve_Concomitant import problem_Concomitant, algo_Concomitant, pathalgo_Concomitant
from classo.solve_Concomitant_Huber import problem_Concomitant_Huber, algo_Concomitant_Huber, pathalgo_Concomitant_Huber
from classo.little_functions import affichage, random_data


'''
Classo and pathlasso are the main functions, they can call every algorithm acording to the method required
'''



def fixlasso(matrix,lam,typ = 'LS', meth='choose',plot_time=True , plot_sol=True,plot_sigm=True , rho = 1.345):
    
        

        
        
    if(typ=='Concomitant'):
        
        t0 = time() 
        
        if not meth in ['ODE','2prox']:
            meth='2prox'
            if (lam>0.1): meth = 'ODE'        # use path algorithm if lam is high, but prox algo if lam is little

        pb = problem_Concomitant(matrix,meth)
        X,s = algo_Concomitant(pb,lam)
        
        dt = time()-t0
    
    
    elif(typ=='Concomitant_Huber'):
        
        t0 = time() 
        
        if not meth in ['2prox']: meth='2prox'

        pb  = problem_Concomitant_Huber(matrix,meth,rho)
        X,s = algo_Concomitant_Huber(pb,lam)
        
        dt = time()-t0
        
    elif(typ=='Huber'):
        
        t0 = time() 
        
        if not meth in ['ODE','2prox','FB','Noproj','cvx']:
            meth='FB'
            if (lam>0.1): meth = 'ODE'        # use path algorithm if lam is high, but prox algo if lam is little

        pb = problem_Huber(matrix,meth,rho)
        X = algo_Huber(pb,lam)
        
        dt = time()-t0     
            
    else: 
       
        t0 = time()  
        
        if not meth in ['ODE','2prox','FB','Noproj','cvx']:
            meth='2prox'
            if (lam>0.1): meth = 'ODE'        # use path algorithm if lam is high, but prox algo if lam is little

        pb = problem_LS(matrix,meth)
        X = algo_LS(pb,lam)
 
        dt = time()-t0
        
    if (plot_sigm and typ in ['Concomitant','Concomitant_Huber']): print('sigma = ',s)
    if (plot_time): print('Running time :', round(dt,5))
    if (plot_sol): plt.bar(range(len(X)),X),plt.title('Problem '+typ+' , lam ='+str(round(lam,3))+' solved with '+ meth +' method'), plt.show()
    if (typ  in ['Concomitant','Concomitant_Huber']): return(X,s)
    return(X)





def pathlasso(matrix,lambdas=False,lamin=1e-2,typ='LS',meth='ODE',plot_time=True,plot_sol=True,plot_sigm=True,rho = 1.345, compare=False, true_lam = False ):

    if (type(compare)!= bool):
        diff,path = compare
        lambdas   = [lam/path[0] for lam in path]
    
    if (type(lambdas)!= bool):
        if (lambdas[0]<lambdas[-1]): lambdas = [lambdas[i] for i in range(len(lambdas)-1,-1,-1)]  # reverse the list needed
        lamin = lambdas[-1]
    else: lambdas = np.linspace(1.,lamin,100)

    if(typ=='Huber'):
        
        t0 = time()
        pb = problem_Huber(matrix,meth,rho)
        lambdamax = pb.lambdamax
        if (true_lam): lambdas=[lamb/lambdamax for lamb in lambdas]
        X  = pathalgo_Huber(pb,lambdas)
        real_path = [lam*lambdamax for lam in lambdas]
        dt = time()-t0
        
        
    elif(typ=='Concomitant'):
        
        t0 = time()
        pb = problem_Concomitant(matrix,meth)
        lambdamax = pb.lambdamax
        if (true_lam): lambdas=[lamb/lambdamax for lamb in lambdas]
        X,S  = pathalgo_Concomitant(pb,lambdas)  
        real_path = [lam*lambdamax for lam in lambdas]
        dt = time()-t0
    
    
    elif(typ=='Concomitant_Huber'):
        
        t0 = time()
        pb = problem_Concomitant_Huber(matrix,meth,rho)
        lambdamax = pb.lambdamax
        if (true_lam): lambdas=[lamb/lambdamax for lamb in lambdas]
        X,S = pathalgo_Concomitant_Huber(pb,lambdas)  
        real_path = [lam*lambdamax for lam in lambdas]
        dt = time()-t0
        
            
    else: 

        t0 = time()
        pb = problem_LS(matrix,meth)
        lambdamax = pb.lambdamax
        if (true_lam): lambdas=[lamb/lambdamax for lamb in lambdas]
        X = pathalgo_LS(pb,lambdas)  
        real_path = [lam*lambdamax for lam in lambdas]
        dt = time()-t0





    if (plot_time): print('Running time :', round(dt,5))
    if (plot_sol): 
        affichage(X,real_path,title=typ+' Path for the method '+meth),plt.show()
        if not (type(compare)==bool): affichage([X[i]-diff[i] for i in range(len(X))],real_path,title='Difference between both methods'),plt.show()
    if (plot_sigm and typ in ['Concomitant','Concomitant Huber']): 
        plt.plot(real_path,S),plt.title('Sigma for Concomitant'),plt.show()
        return(X,real_path,S)
    return(X,real_path)
 
    
    
    
    


# Cost fucntions for the three 'easiest' problems. Useful for test, to compare two solutions slightly different
def L_LS(A,y,lamb,x): return(LA.norm( A.dot(x) - y )**2 + lamb * LA.norm(x,1))
def L_conc(A,y,lamb,x): return(LA.norm( A.dot(x) - y ) + np.sqrt(2)*lamb * LA.norm(y,1))
def L_H(A,y,lamb,x,rho): return(hub( A.dot(x) - y , rho) + lamb * LA.norm(x,1))

def hub(r,rho) : 
    h=0
    for j in range(len(r)):
        if(abs(r[j])<rho): h+=r[j]**2
        elif(r[j]>0)     : h+= (2*r[j]-rho)*rho
        else             : h+= (-2*r[j]-rho)*rho
    return(h)




def ordre(meth):
    m,r_nonzero,k,sigma = 100,0.1,4,0.5
    D,T = range(50,500,20),[]
    for d in D:    
        s=0
        N_moy = d//10
        for i in range(N_moy):
            matrices,sol = random_data(m,d,int(d*r_nonzero),k,sigma)
            t0=time()
            X = Classo(matrices,0.5,meth=meth,plot_sol=False,plot_time=False)
            s+=time()-t0
            constr = LA.norm(matrices[1].dot(X))
            if (constr > 1e-10): print('C')
        T.append(s/N_moy)
    return(D,T)


import numpy as np
import numpy.linalg as LA
import scipy.io as sio


import Functions.functions_4algos as algoF


def generate_random(dim):                # Function to generate random A, y, and C
        (m,d,k,d_nonzero,sigma) = dim
        A, sol, sol_reduc= np.random.randn(m, d),np.zeros(d), np.random.rand(d_nonzero)
        if (k==0):
            C , list_i = np.zeros((1,d)), np.random.randint(d, size=d_nonzero)
            sol[list_i]=sol_reduc
        else:
            rank1,rank2 = 0,0
            while (rank1 !=k) :        
                C = np.random.randint(low=-1,high=2, size=(k, d))
                rank1 = LA.matrix_rank(C)
            while (rank2!=k):
                list_i = np.random.randint(d, size=d_nonzero)
                C_reduc = np.array([C.T[i] for i in list_i]).T
                rank2 = LA.matrix_rank(C_reduc)
            proj = sol_reduc - (C_reduc.T).dot(algoF.proj_c(C_reduc).dot(sol_reduc))
            sol[list_i]=proj
            return(A,C,sol)



class problem :
    
    def __init__(self,data,ty,normalise=True):
        self.N = 500000
        
        if(len(data)==3):self.matrix, self.dim = data, (data[0].shape[0],data[0].shape[1],data[1].shape[0])
        
        elif(len(data)==5):
            (A,C,sol), self.dim = generate_random(data), data[:3]
            self.sol,y = sol, A.dot(sol)+np.random.randn(data[0])*data[-1]
            self.matrix = (A,C,y)
        
        (m,d,k) = self.dim
        self.weights = np.ones(d)
        self.init = 1.,1.,np.zeros(m), np.zeros(d), np.zeros(d), np.zeros(k)
        self.tol = 1e-7
         
        
        algo = ty
        self.regpath = False
        self.name = ty
        self.type = algo          # type of algorithm used
        
        self.prox_phi = algoF.prox_phi_1
        self.mu = 1.95
        
        
        
        self.r = 1             # parameter for Huber problem : LS + lambda |b|_1 + lambda/r |o|_1
        
        self.gam = 1.
        self.tau = 0.5         # equation for the convergence of Noproj and LS algorithms : 2.gam + tau < 1
        if (ty=='LSG' or ty == 'Concomitant'): self.gam = self.dim[1]**(1.5)
       
        if normalise: self.normalize()
        else: self.My, self.lnA,self.lny = 0., [1.]*d, 1.
            
        self.c = (d/LA.norm(self.matrix[0],2))**2  # parameter for Concomitant problem : the matrix is scaled as c*A^2 
        self.compute_param()
        
            
    def compute_param(self):
        (A,C,y) = self.matrix
        m,d,k = self.dim
        coef = self.c
        Q1 = coef*(A.T).dot(LA.inv(2*np.eye(A.shape[0])+coef*A.dot(A.T)))
        Q2 = LA.inv(2*np.eye(A.shape[1])+coef*(A.T).dot(A))
        self.Q = (Q1, Q2)
        if (LA.matrix_rank(C)==0):  self.projC = C
        else:     self.projC = np.linalg.inv(C.dot(C.T) ).dot( C )
        self.AtA =(A.T).dot(A)
        self.Aty = (A.T).dot(y)
        self.Cnorm  = LA.norm(C,2)**2
        self.tauN= self.tau/self.Cnorm
        self.lambdamax1 = 2*LA.norm((A.T).dot(y),np.infty)
        self.lambdamax = self.lambdamax1
        self.AtAnorm = LA.norm(self.AtA,2)
        
        
        
    def normalize(self):
        (A,C,y),(m,d,k) = self.matrix, self.dim
        self.My = sum(y)/m
        self.lnA = [LA.norm(A[:,j]) for j in range(d)]
        self.lny = LA.norm(y-self.My*np.ones(m))
        la,ly,my = self.lnA, self.lny , self.My
        An = np.array([[A[i,j]/(la[j])  for j in range(d)] for i in range(m)])
        yn = np.array([ (y[j]-my)/ly for j in range(m) ])
        Cn = np.array([[C[i,j]*ly/la[j] for j in range(d)] for i in range(k)])
        self.matrix = (An,Cn,yn)
        
    def denormalize(self):
        (A,C,y),(m,d,k),la,ly,my = self.matrix, self.dim, self.lnA, self.lny , self.My
        An = np.array([[A[i,j]*la[j]  for j in range(d)] for i in range(m)])
        Cn = np.array([[C[i,j]*la[j]/ly  for j in range(d)] for i in range(k)])
        yn = np.array([ (y[j]+my)*ly for j in range(m) ])
        self.matrix = (An,Cn,yn)  
        
        
    def change_type(self,typ):
        self.type,self.name = typ,typ
        self.gam , self.tauN  = .5 , .1/self.Cnorm
        if (typ=='LSG' or typ == 'Concomitant'): self.gam = np.sqrt(self.dim[1])
        if (typ == 'Concomitant'): self.lambdamax = self.lambdamax1 / np.sqrt(2)

            
            
            
            
            
            
            


from setuptools import setup, find_packages

setup(name='c_lasso',
      version='0.1',
      license='MIT',
      author='Leo Simpson',
      url='https://github.com/Leo-Simpson/CLasso',
      author_email='leo.bill.simpson@gmail.com',
      description='Algorithms for constrained Lasso problems',
      packages=['classo'],
      long_description=open('README.md').read(),
      zip_safe=False)


import numpy as np
import matplotlib.pyplot as plt
import numpy.random as rd
import numpy.linalg as LA
from classo.compact_func import fixlasso,pathlasso
n_lam = 100

def train_test_CV(n,k,test_pourcent):
    idx, training_size = rd.permutation(n), int(n-n*test_pourcent)
    idx_train, idx_test = idx[:training_size], idx[training_size:]
    SUBLIST,end = [],0 
    for i in range(k):
        begin,end = end,end+training_size//k
        if(i<training_size%k): end+=1
        SUBLIST.append(idx[begin:end])
    return(SUBLIST, idx_train,idx_test) 


def train_test_i (SUBLIST,i):
    training_set,test_set = np.array([],dtype=int),SUBLIST[i]
    for j in range(len(SUBLIST)):
        if (j != i): training_set = np.concatenate((training_set,SUBLIST[j]))
    return(training_set,test_set)
            

def training(matrices,typ,lamin, training_set):
    (A,C,y)   = matrices
    mat       = (A[training_set],C,y[training_set]) 
    return(pathlasso(mat,lamin=lamin,typ=typ,meth='ODE',plot_time=False,plot_sol=False,plot_sigm=False)[0])


def test_i (matrices,typ,lamin, SUBLIST,i):
    training_set,test_set = train_test_i (SUBLIST,i)
    BETA                  = training(matrices,typ,lamin, training_set)
    L = []
    for j in range(n_lam):
        L.append(accuracy_func(matrices[0][test_set],matrices[2][test_set],BETA[j],typ))
    return(L)

def average_test(matrices,typ,lamin, SUBLIST):
    AVG = [0]*n_lam
    for i in range(len(SUBLIST)):
        L = test_i (matrices,typ,lamin, SUBLIST,i)
        for j in range(n_lam):
            AVG[j]+=L[j]
    return(AVG)

def CV(matrices,k=5,typ='LS',test=0.4,lamin=1e-2, print_lam= True):
    (A,C,y) = matrices
    SUBLIST, idx_train, idx_test = train_test_CV(len(y),k,test)
    AVG1  = average_test(matrices,typ,lamin, SUBLIST) 
    LAM = np.linspace(1.,lamin,n_lam)
    Davg = (AVG1[2]-AVG1[-1])
    AVG2 = [AVG1[i] - LAM[i]*Davg for i in range(n_lam)]
    i1 = np.argmin(AVG1)
    i2 = np.argmin(AVG2)
    lam1 = LAM[i1]
    lam2 = LAM[i2]
    if(print_lam): print('lam',lam2)
    beta1 = fixlasso((A[idx_train],C,y[idx_train]),lam1,typ=typ,plot_time=False,plot_sol=False,plot_sigm=False)
    beta2 = fixlasso((A[idx_train],C,y[idx_train]),lam2,typ=typ,plot_time=False,plot_sol=False,plot_sigm=False)
    return(beta2)
    
        
# Cost fucntions for the three 'easiest' problems. 
def hub(r,rho) : 
    h=0
    for j in range(len(r)):
        if(abs(r[j])<rho): h+=r[j]**2
        elif(r[j]>0)     : h+= (2*r[j]-rho)*rho
        else             : h+= (-2*r[j]-rho)*rho
    return(h)

def accuracy_func(A,y,beta, typ='LS',rho = 1.345):      
    if (typ == 'Huber'):    return(hub( A.dot(beta) - y , rho * LA.norm(y,np.infty)/np.sqrt(len(y))) )
    else :                  return(LA.norm( A.dot(beta) - y )**2)


import numpy as np
import numpy.linalg as LA
    
'''    
Problem    :   min h_rho((Ab - y)/sigma)sigma + simga + lambda ||b||1 with C.b= 0, sigma>0 

Dimensions :   A : m*d  ;  y : m  ;  b : d   ; C : k*d

The first function compute a solution of a Lasso problem for a given lambda. The parameters are lam (lambda/lambdamax, \in [0,1]) and pb, which has to be a 'problem_LS type', which is defined bellow in order to contain all the important parameters of the problem. One can initialise it this way : pb = class_of_problem.problem(data=(A,C,y),type_of_algo). We solve the problem without normalizing anything. 
'''    


def algo_Concomitant_Huber(pb,lam):
    
    pb_type = pb.type      # 2prox
    (m,d,k),(A,C,y)  = pb.dim,pb.matrix 
    lamb,rho  = lam * pb.lambdamax, pb.rho

    regpath = pb.regpath
    Anorm   = LA.norm(A,'fro')
    tol     = pb.tol * LA.norm(y) # tolerance rescaled
    Proj    = proj_c(C,d)   # Proj = I - C^t . (C . C^t )^-1 . C 
    gamma   = LA.norm(y)*pb.gam / (Anorm**2)
    w,zerod = lamb *gamma*pb.weights, np.zeros(d) # two vectors usefull to compute the prox of f(b)= sum(wi |bi|)
    mu, c   = pb.mu, pb.c
    Q1,Q2   = QQ(c,A) 
    QA,qy   = Q1.dot(A), Q1.dot(y)
    root    = [0.]*len(y)
    proj_sigm = lambda vect: ([sum(vect)/len(vect)]*len(vect))
    xs,nu,o,xbar,x = pb.init
    #2prox
    print(pb.sigmax)
    if (pb_type == '2prox'):
        
        for i in range(pb.N):
            nv_b, nv_s = x + Q1.dot(o) - QA.dot(x) - Q2.dot(x-xbar), (xs+nu)/2
            if (i>0 and LA.norm(b-nv_b)*Anorm +LA.norm(s-nv_s)<2*tol):
                if (regpath):return(b,(xs,nu,o,xbar,x),sum(s)/len(s)/pb.sigmax) 
                else :       return(b,sum(s)/len(s))
                        
            s,b = nv_s, nv_b
            Ab = A.dot(b)
            p1,p2,root = prox_phi_2(xs,2*Ab-o-y,gamma/c,root,rho)
            sup = [ proj_sigm(nu)-s , p1-s , p2 + y - Ab , prox(2*b-xbar,w,zerod)-b, Proj.dot(2*b-x)-b]
            xs,nu,o,xbar,x = xs+mu*sup[0] ,  nu+mu*sup[1] ,  o+mu*sup[2] ,  xbar+mu*sup[3] ,  x+mu*sup[4]
            if (LA.norm(b)+LA.norm(s)>1e6): 
                print('DIVERGENCE')
                return(b,np.sqrt(m)*sum(s)/len(s))
        print('NO CONVERGENCE')
        return(b,sum(s)/len(s))
    
    
    
    
    
    print('none of the cases ! ')        
    
    

    
'''
This function compute the the solution for a given path of lam : by calling the function 'algo' for each lambda with warm start, or wuth the method ODE, by computing the whole path thanks to the ODE that rules Beta and the subgradient s, and then to evaluate it in the given finite path.  
'''
    
def pathalgo_Concomitant_Huber(pb,path):
    n = pb.dim[0]
    BETA,SIGMA,tol = [],[],pb.tol

    save_init = pb.init   
    pb.regpath = True
    for lam in path:
        X = algo_Concomitant_Huber(pb,lam)     
        BETA.append(X[0]), SIGMA.append(X[2])
        pb.init = X[1]    
        if(sum([ (abs(X[0][i])>1e-2) for i in range(len(X[0])) ])>=n):
                pb.init, BETA = save_init, BETA + [BETA[-1]]*(len(path)-len(BETA))
                print('stop the path because number of active param reach n')
                return(BETA,SIGMA)
            
    pb.init = save_init
    pb.regpath = False
    return(BETA,SIGMA)








'''
Class of problem : we define a type, which will contain as keys, all the parameters we need for a given problem.
'''


class problem_Concomitant_Huber :
    
    def __init__(self,data,algo,rho):
        self.N = 500000
        
        if(len(data)==3): (A,C,y), self.dim = data, (data[0].shape[0],data[0].shape[1],data[1].shape[0])
        elif(len(data)==5):
            (A,C,sol), self.dim = generate_random(data), data[:3]
            self.sol,y = sol, A.dot(sol)+np.random.randn(data[0])*data[-1]
        self.matrix = (A,C,y)
        
        (m,d,k) = self.dim
        self.weights = np.ones(d)
        self.tol = 1e-6
         
          
        self.regpath = False
        self.name = algo + ' Concomitant Huber'
        self.type = algo          # type of algorithm used

        self.rho = rho
        self.mu  = 1.95
         
        self.c = (d/LA.norm(A,2))**2  # parameter for Concomitant problem : the matrix is scaled as c*A^2 
        self.gam = np.sqrt(d)
        sigmax = find_sigmax(y,rho)
        self.lambdamax = 2/sigmax*LA.norm((A.T).dot(h_prime(y,rho*sigmax)),np.infty)
        self.init = sigmax*np.ones(m),sigmax*np.ones(m),np.zeros(m), np.zeros(d), np.zeros(d)
           

        



'''
Functions used in the algorithms, modules needed : 
import numpy as np
import numpy.linalg as LA
from .../class_of_problem import problem
'''


# compute the prox of the function : f(b)= sum (wi * |bi| )
def prox(b,w,zeros): return(np.minimum(b+w,zeros)+np.maximum(b-w,zeros)) 

# Compute I - C^t (C.C^t)^-1 . C : the projection on Ker(C)
def proj_c(M,d):
    if (LA.matrix_rank(M)==0):  return(np.eye(d))
    return(np.eye(d)-LA.multi_dot([M.T,np.linalg.inv(M.dot(M.T) ),M]) )



def QQ(coef,A): return(coef*(A.T).dot(LA.inv(2*np.eye(A.shape[0])+coef*A.dot(A.T))),LA.inv(2*np.eye(A.shape[1])+coef*(A.T).dot(A)))    



# Compute the real positive root of a polynomial of degree 3 in the form : X^3 + a*X - b with Newton method and a warm start (for Comcomitant problem)
def calc_Newton(a,b,root):
    er = -b
    while (abs(er)>1e-6): 
        root= root-er/(3*root**2+a)
        er = root**3 + a*root-b
    return(root)



    
def prox_phi_2(sig,u,gamma,warm_start,rho):
    p,q, ws = np.zeros(len(u)), np.zeros(len(u)),  np.zeros(len(u))
    for i in range(len(u)):
        p[i],q[i],ws[i]=prox_phi_i(sig[i],u[i],2*gamma,warm_start[i],rho)
    return(p,q,ws)


def prox_phi_i(s,u,gamma,root,rho):
    if (u==0.): return(0,0,root)
    frac = gamma*rho/abs(u)
    term = s+gamma*(rho**2-1)/2
    bool1 = (frac>=1) 
    bool2 = (abs(u)**2<=gamma*(gamma-2*s))
    bool3 = (term<=0)
    bool4 = (abs(u)>=rho*s + gamma*rho*(1+rho**2)/2)
    
    
    if   (bool1  and bool2 ): return(0.,0.,root)
    elif (bool3 and not bool1): return(0.,u*(1-frac),root) 
    elif (not bool3 and bool4):  return(term,u*(1-frac),root)
        
    root = calc_Newton(2*s/gamma+1,2*abs(u)/gamma,root)
    return(s+gamma*(root**2-1)/2,u-gamma*root*np.sign(u),root)


# Compute the derivative of the huber function, particulary useful for the computing of lambdamax 
def h_prime(y,rho):
    m = len(y)
    lrho = rho*np.ones(m)
    return(np.maximum(lrho,-y)+ np.minimum(y-lrho,0))




def find_sigmax(y,rho):
    m,evol = len(y), True
    F = [True]*m
    if (rho > 1):
        while(evol):
            evol = False
            t = np.sqrt(m-(m-sum(F))*rho**2)/LA.norm(y[F])
            cste= rho/t
            for j in range(m):
                if (F[j] and y[j]>cste): F[j],evol= False , True
        return(1/t)
    else:
        print('rho too little ==> sigma is always 0')



def generate_random(dim):                # Function to generate random A, y, and C with given dimensions 
        (m,d,k,d_nonzero,sigma) = dim
        A, sol, sol_reduc= np.random.randn(m, d),np.zeros(d), np.random.rand(d_nonzero)
        if (k==0):
            C , list_i = np.zeros((1,d)), np.random.randint(d, size=d_nonzero)
            sol[list_i]=sol_reduc
        else:
            rank1,rank2 = 0,0
            while (rank1 !=k) :        
                C = np.random.randint(low=-1,high=2, size=(k, d))
                rank1 = LA.matrix_rank(C)
            while (rank2!=k):
                list_i = np.random.randint(d, size=d_nonzero)
                C_reduc = np.array([C.T[i] for i in list_i]).T
                rank2 = LA.matrix_rank(C_reduc)
            proj = proj_c(C_reduc,d_nonzero).dot(sol_reduc)
            sol[list_i]=proj
            return(A,C,sol)

